@@ -6,62 +6,23 @@
 #  http://www.upnp.org/schemas/av/didl-lite-v2.xsd
 #  http://xml.coverpages.org/mpeg21-didl.html
 
-import re
-
-from typing import Any, Dict, List, Optional  # noqa: F401 pylint: disable=unused-import
+from typing import Any, Dict  # noqa: F401 pylint: disable=unused-import
+from typing import cast, List, Optional, Sequence, Type, TypeVar, Union
 from xml.etree import ElementTree as ET
 
+import defusedxml.ElementTree  # type: ignore
 
-NAMESPACES = {
-    'didl_lite': 'urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/',
-    'dc': 'http://purl.org/dc/elements/1.1/',
-    'upnp': 'urn:schemas-upnp-org:metadata-1-0/upnp/',
-    'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
-}
+from .utils import NAMESPACES
+from .utils import didl_property_def_key
+from .utils import namespace_tag
+from .utils import ns_tag
+from .utils import to_camel_case
 
 
-def _ns_tag(tag: str) -> str:
-    """
-    Expand namespace-alias to url.
-
-    E.g.,
-        _ns_tag('didl_lite:item') -> '{urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/}item'
-    """
-    if ':' not in tag:
-        return tag
-
-    namespace, tag = tag.split(':')
-    namespace_uri = NAMESPACES[namespace]
-    return '{{{0}}}{1}'.format(namespace_uri, tag)
-
-
-def _namespace_tag(namespaced_tag: str) -> str:
-    """
-    Extract namespace and tag from namespaced-tag.
-
-    E.g., _namespace_tag('{urn:schemas-upnp-org:metadata-1-0/upnp/}class') ->
-        'urn:schemas-upnp-org:metadata-1-0/upnp/', 'class'
-    """
-    if '}' not in namespaced_tag:
-        return None, namespaced_tag
-
-    idx = namespaced_tag.index('}')
-    namespace = namespaced_tag[1:idx]
-    tag = namespaced_tag[idx + 1:]
-    return namespace, tag
-
-
-def _to_camel_case(name):
-    sub1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
-    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', sub1).lower()
-
-
-def _didl_property_def_key(didl_property_def):
-    """Get Python property key for didl_property_def."""
-    if didl_property_def[1].startswith('@'):
-        return _to_camel_case(didl_property_def[1].replace('@', ''))
-
-    return _to_camel_case(didl_property_def[1].replace('@', '_'))
+TDO = TypeVar('TDO', bound='DidlObject')
+TC = TypeVar('TC', bound='Container')
+TD = TypeVar('TD', bound='Descriptor')
+TR = TypeVar('TR', bound='Resource')
 
 
 # region: DidlObjects
@@ -81,7 +42,8 @@ class DidlObject:
         ('upnp', 'writeStatus', 'O'),
     ]
 
-    def __init__(self, id="", parent_id="", descriptors=None, **properties):
+    def __init__(self, id: str = "", parent_id: str = "",
+                 descriptors: Optional[Sequence['Descriptor']] = None, **properties: Any):
         """Initializer."""
         # pylint: disable=invalid-name,redefined-builtin
         properties['id'] = id
@@ -93,25 +55,25 @@ class DidlObject:
         self.resources = properties.get('resources') or []
         self.descriptors = descriptors if descriptors else []
 
-    def _ensure_required_properties(self, **properties):
+    def _ensure_required_properties(self, **properties: Any) -> None:
         """Check if all required properties are given."""
         for property_def in self.didl_properties_defs:
-            key = _didl_property_def_key(property_def)
+            key = didl_property_def_key(property_def)
             if property_def[2] == 'R' and key not in properties:
                 raise Exception(key + ' is mandatory')
 
-    def _set_properties(self, **properties):
+    def _set_properties(self, **properties: Any) -> None:
         """Set attributes from properties."""
         # ensure we have default/known slots
         for property_def in self.didl_properties_defs:
-            key = _didl_property_def_key(property_def)
+            key = didl_property_def_key(property_def)
             setattr(self, key, None)
 
         for key, value in properties.items():
             setattr(self, key, value)
 
     @classmethod
-    def from_xml(cls, xml_el: ET.Element):
+    def from_xml(cls: Type[TDO], xml_el: ET.Element) -> TDO:
         """
         Initialize from an XML node.
 
@@ -122,23 +84,23 @@ class DidlObject:
 
         # attributes
         for attr_key, attr_value in xml_el.attrib.items():
-            key = _to_camel_case(attr_key)
+            key = to_camel_case(attr_key)
             properties[key] = attr_value
 
         # child-nodes
         for xml_child_node in xml_el:
-            if xml_child_node.tag == _ns_tag('didl_lite:res'):
+            if xml_child_node.tag == ns_tag('didl_lite:res'):
                 continue
 
-            _, tag = _namespace_tag(xml_child_node.tag)
-            key = _to_camel_case(tag)
+            _, tag = namespace_tag(xml_child_node.tag)
+            key = to_camel_case(tag)
             value = xml_child_node.text
             properties[key] = value
 
             # attributes of child nodes
             parent_key = key
             for attr_key, attr_value in xml_child_node.attrib.items():
-                key = parent_key + '_' + _to_camel_case(attr_key)
+                key = parent_key + '_' + to_camel_case(attr_key)
                 properties[key] = attr_value
 
         # resources
@@ -158,21 +120,22 @@ class DidlObject:
 
     def to_xml(self) -> ET.Element:
         """Convert self to XML Element."""
-        item_el = ET.Element(_ns_tag(self.tag))
+        assert self.tag is not None
+        item_el = ET.Element(ns_tag(self.tag))
         elements = {'': item_el}
 
         # properties
         for property_def in self.didl_properties_defs:
             if '@' in property_def[1]:
                 continue
-            key = _didl_property_def_key(property_def)
+            key = didl_property_def_key(property_def)
 
             if getattr(self, key) is None or \
                key == 'res':  # no resources, handled later on
                 continue
 
             tag = property_def[0] + ':' + property_def[1]
-            property_el = ET.Element(_ns_tag(tag), {})
+            property_el = ET.Element(ns_tag(tag), {})
             property_el.text = getattr(self, key)
             item_el.append(property_el)
             elements[property_def[1]] = property_el
@@ -182,7 +145,7 @@ class DidlObject:
             if '@' not in property_def[1]:
                 continue
 
-            key = _didl_property_def_key(property_def)
+            key = didl_property_def_key(property_def)
             value = getattr(self, key)
             if value is None:
                 continue
@@ -203,6 +166,12 @@ class DidlObject:
 
         return item_el
 
+    def __getattr__(self, name: str) -> Any:
+        """Get attribute."""
+        if name not in self.__dict__:
+            raise AttributeError(name)
+        return self.__dict__[name]
+
 
 # region: items
 class Item(DidlObject):
@@ -556,7 +525,7 @@ class Container(DidlObject, list):
     ]
 
     @classmethod
-    def from_xml(cls, xml_el: ET.Element):
+    def from_xml(cls: Type[TC], xml_el: ET.Element) -> TC:
         """
         Initialize from an XML node.
 
@@ -828,9 +797,12 @@ class Resource:
 
     # pylint: disable=too-few-public-methods,too-many-instance-attributes
 
-    def __init__(self, uri, protocol_info, import_uri=None, size=None, duration=None,
-                 bitrate=None, sample_frequency=None, bits_per_sample=None,
-                 nr_audio_channels=None, resolution=None, color_depth=None, protection=None):
+    def __init__(self, uri: Optional[str], protocol_info: str, import_uri: Optional[str] = None,
+                 size: Optional[str] = None, duration: Optional[str] = None,
+                 bitrate: Optional[str] = None, sample_frequency: Optional[str] = None,
+                 bits_per_sample: Optional[str] = None, nr_audio_channels: Optional[str] = None,
+                 resolution: Optional[str] = None, color_depth: Optional[str] = None,
+                 protection: Optional[str] = None):
         """Initializer."""
         # pylint: disable=too-many-arguments
         self.uri = uri
@@ -847,7 +819,7 @@ class Resource:
         self.protection = protection
 
     @classmethod
-    def from_xml(cls, xml_node: ET.Element):
+    def from_xml(cls: Type[TR], xml_node: ET.Element) -> TR:
         """Initialize from an XML node."""
         uri = xml_node.text
         protocol_info = xml_node.attrib["protocolInfo"]
@@ -874,7 +846,7 @@ class Resource:
         attribs = {
             'protocolInfo': self.protocol_info,
         }
-        res_el = ET.Element(_ns_tag('res'), attribs)
+        res_el = ET.Element(ns_tag('res'), attribs)
         res_el.text = self.uri
         return res_el
 
@@ -882,7 +854,8 @@ class Resource:
 class Descriptor:
     """DIDL Descriptor."""
 
-    def __init__(self, id, name_space, type=None, text=None):
+    def __init__(self, id: str, name_space: str, type: Optional[str] = None,
+                 text: Optional[str] = None):
         """Initializer."""
         # pylint: disable=invalid-name,redefined-builtin
         self.id = id
@@ -891,7 +864,7 @@ class Descriptor:
         self.text = text
 
     @classmethod
-    def from_xml(cls, xml_node: ET.Element):
+    def from_xml(cls: Type[TD], xml_node: ET.Element) -> TD:
         """Initialize from an XML node."""
         id_ = xml_node.attrib['id']
         name_space = xml_node.attrib['nameSpace']
@@ -903,40 +876,47 @@ class Descriptor:
         attribs = {
             'id': self.id,
             'nameSpace': self.name_space,
-            'type': self.type,
         }
-        desc_el = ET.Element(_ns_tag('desc'), attribs)
+        if self.type is not None:
+            attribs['type'] = self.type
+        desc_el = ET.Element(ns_tag('desc'), attribs)
         desc_el.text = self.text
         return desc_el
+
+    def __getattr__(self, name: str) -> Any:
+        """Get attribute."""
+        if name not in self.__dict__:
+            raise AttributeError(name)
+        return self.__dict__[name]
 # endregion
 
 
-def to_xml_string(*objects) -> str:
+def to_xml_string(*objects: DidlObject) -> bytes:
     """Convert items to DIDL-Lite XML string."""
-    root_el = ET.Element(_ns_tag('DIDL-Lite'), {})
+    root_el = ET.Element(ns_tag('DIDL-Lite'), {})
     root_el.attrib['xmlns'] = NAMESPACES['didl_lite']
 
     for didl_object in objects:
         didl_object_el = didl_object.to_xml()
         root_el.append(didl_object_el)
 
-    return ET.tostring(root_el)
+    return cast(bytes, ET.tostring(root_el))
 
 
-def from_xml_string(xml_string) -> List[DidlObject]:
+def from_xml_string(xml_string: str) -> List[Union[DidlObject, Descriptor]]:
     """Convert XML string to DIDL Objects."""
-    xml_el = ET.fromstring(xml_string)
+    xml_el = defusedxml.ElementTree.fromstring(xml_string)
     return from_xml_el(xml_el)
 
 
-def from_xml_el(xml_el: ET.Element) -> List[DidlObject]:
+def from_xml_el(xml_el: ET.Element) -> List[Union[DidlObject, Descriptor]]:
     """Convert XML Element to DIDL Objects."""
-    didl_objects = []
+    didl_objects = []  # type: List[Union[DidlObject, Descriptor]]
 
     # items and containers, in order
     for child_el in xml_el:
-        if child_el.tag != _ns_tag('didl_lite:item') and \
-           child_el.tag != _ns_tag('didl_lite:container'):
+        if child_el.tag != ns_tag('didl_lite:item') and \
+           child_el.tag != ns_tag('didl_lite:container'):
             continue
 
         # construct item
@@ -944,6 +924,8 @@ def from_xml_el(xml_el: ET.Element) -> List[DidlObject]:
         if upnp_class is None or not upnp_class.text:
             continue
         didl_object_type = type_by_upnp_class(upnp_class.text)
+        if didl_object_type is None:
+            continue
         didl_object = didl_object_type.from_xml(child_el)
         didl_objects.append(didl_object)
 
@@ -956,7 +938,7 @@ def from_xml_el(xml_el: ET.Element) -> List[DidlObject]:
 
 
 # upnp_class to python type mapping
-def type_by_upnp_class(upnp_class: str) -> type:
+def type_by_upnp_class(upnp_class: str) -> Optional[Type[DidlObject]]:
     """Get DidlObject-type by upnp_class."""
     queue = DidlObject.__subclasses__()
     while queue:
@@ -965,3 +947,4 @@ def type_by_upnp_class(upnp_class: str) -> type:
 
         if type_.upnp_class == upnp_class:
             return type_
+    return None
